function p=isc(p,alvin,varargin)
% isc: compute canonical paths to a CSS or CPS, continuation in alpha or arclength
% 
% wrapper for cpsolver. 
% Computes a canonical path from p.ocopt.s0 to a CPS/CSS p.ocopt.s1 with sattle
% point property by continuation in the initial state either via natural
% continuation or (pseudo) arclength continuation.
% p: Problem strucure initialized by ocinit, see this for details
% alvin: vector of continuation steps in alpha, i.e. initial solution for the
% canonical path will be al*IS+(1-al)*FS, where IS is the aimed initial
% solution, FS is the CSS/CPS and al is the current alpha value. Thus alvin
% should start small and end in 1, e.g. alvin=0:0.1:1
% nsteps: (optional) number of arclength steps after the last value in alvin
% is reached. Continuation is automatically aborted if alpha is close to
% one and a last continuation in natural parametrization is made to reach 1
% exactly.

if p.ocopt.start==1 % startup
    if p.ocopt.s1ho==1
        if ~isfield(p.ocopt,'F')
            [p.ocopt.F, p.ocopt.F2,p.ocopt.muv]=floqpsmatadj(p.ocopt.s1); % Computes matrix of adjoint eigenvectors to unstable/not stable eigenvalues, i.e. kerF projects on stable eigenspace
            [~,R]=qr(p.ocopt.F2);
            p.ocopt.F2=real(R);
            if max(abs(imag(R(:))))>1e-8
                warning('Projection to center unstable eigenspace has no real basis. Computations can be false.');
            end
            fprintf('%g \n',abs(p.ocopt.muv));pause
        end
    else
        [p.ocopt.F,p.ocopt.muv,~,T]=getPsi2(p.ocopt.s1); % unstable directions for CSS
        if isempty(p.ocopt.T)
            p.ocopt.T=T; % Timeguess by getPsi - computed via smallest eigenvector
        end
    end
    if ~isfield(p.ocopt,'tv') || isempty(p.ocopt.tv) % set basic t-grid
        if p.ocopt.s1ho==1
            p.ocopt.tv=linspace(0,1,p.ocopt.nti);
        else
            p.ocopt.tv=linspace(0,1,p.ocopt.nti).^2;
        end
    end
    p.tomopt.M=p.ocopt.s1.mat.M; % restore mass matrix for newtom
    p.ocopt.freeT=0; % Endtime T is fixed, will be free in later continuations
    p.hist.alpha=[]; % vector of alpha-values
    p.hist.vv=[]; % vector of objective values
    p.hist.tt=cell(0,0); % cell of time vectors at each continuation step
    p.hist.u=cell(0,0); % cell of u vectors at each continuation step
    p.hist.par=cell(0,0); % cell of par vectors at each continuation step    
end % startup finished

% initialise some stuff
p.ocopt.start=0; % skip startup for repeated calls
n=p.ocopt.s0.nu+p.ocopt.s0.nc.nq; % number of unknowns
if ~isempty(alvin)
    al=alvin(1); % first continuation
    if p.ocopt.s1ho==1 % starting/end sol for CPS resp. CSS
        p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.hopf.y(1:n,end-1); % start point
        initpar=p.ocopt.wT*p.ocopt.s1.hopf.T;
        p.ocopt.T=initpar;
        p.ocopt.u1=p.ocopt.s1.hopf.y(1:n,end-1); % end point
    else
        initpar=p.ocopt.T;
        if ~isfield(p.ocopt,'lastT');p.ocopt.lastT=p.ocopt.T;end
        p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.u(1:n); % start point
        p.ocopt.u1=p.ocopt.s1.u(1:n); % end point
    end
    if isempty(p.cp) % construct initial guess if not already given
        if p.ocopt.s1ho==1
            p.cp.t=p.ocopt.tv;
            p.cp.par=initpar;
            % creates round(p.ocopt.wT) periods of the hopf orbit as starting
            % solution
            pT=round(p.ocopt.wT);
            pnt=length(p.ocopt.s1.hopf.t)-1;
            repf=repelem(1/pT,(pT-1)*pnt);
            repa=repelem(1:pT-1,pnt);
            ht=[1/pT*p.ocopt.s1.hopf.t (repmat(p.ocopt.s1.hopf.t(2:end),1,pT-1)+repa).*repf];
            hu=[p.ocopt.s1.hopf.y(1:n,:)';repmat(p.ocopt.s1.hopf.y(1:n,2:end)',pT-1,1)];
            p.cp.u=(interp1(ht,hu,p.cp.t))';
        else
            p.cp.t=p.ocopt.tv;
            p.cp.u=p.ocopt.u0*ones(1,length(p.ocopt.tv))+(p.ocopt.u1-p.ocopt.u0)*p.ocopt.tv;
            p.cp.par=initpar;
        end
    end
    if ~isfield(p.ocopt,'usec') || isempty(p.ocopt.usec) || p.ocopt.msw==0 % compute sol for first al if no secant given or trivial pred      
        switch p.ocopt.verb; 
            case 1; fprintf('alpha=%g\n ',al); 
            case 2;    fprintf('alpha=%g - Starting computation...\n',al);
        end
        [sol1,p.ocopt.s1]=cpsolver(p.cp,p,0);
        if sol1.err~=0
            warning('Did not find any solution in first continuation-step, error-code: %i. \n Last solution in p.cp.',sol1.err);
            p.cp=sol1;
            return;
        end
        if p.ocopt.retsw==1
            p.hist.tt=[p.hist.tt,sol1.t];
            p.hist.u=[p.hist.u,sol1.u];
            p.hist.par=[p.hist.par,sol1.par];
        else
            p.hist.tt={sol1.t};
        end
        p.hist.alpha=[p.hist.alpha, al];
        val=jcaiT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi))+disjcaT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi),p.ocopt.s1ho);
        p.hist.vv=[p.hist.vv val];
        p=errcheck(p,sol1);
        p.ocopt.lastT=p.cp.par(1);
        p.ocopt.um2=p.cp.u;
        p.ocopt.parm2=p.cp.par;
        if p.ocopt.msw==0
            mst=length(alvin);
        else
            mst=min(2,length(alvin));
        end
        for i=[2:mst] % remaining steps with trivial predictor, resp. 2nd step for secant
            al=alvin(i);
            if p.ocopt.s1ho==1
                p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.hopf.y(1:n,end-1); % start point for alvin(i);
            else
                p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.u(1:n); % start point for alvin(i);
            end
            switch p.ocopt.verb; 
                case 1; fprintf('alpha=%g\n',al);
                case 2; fprintf('alpha=%g - Starting computation...\n',al);
            end
            told=p.cp.t;
            [sol1,p.ocopt.s1]=cpsolver(p.cp,p,0);
            if sol1.err~=0
                warning('Did not find any solution in %i continuation-step, error-code: %i. \n Last solution in p.cp.',i,sol1.err);
                p.cp=sol1;
                return;
            end
            if p.ocopt.retsw==1
                p.hist.tt=[p.hist.tt,sol1.t];
                p.hist.u=[p.hist.u,sol1.u];
                p.hist.par=[p.hist.par,sol1.par];
            else
                p.hist.tt={sol1.t};
            end
            p.hist.alpha=[p.hist.alpha al];
            val=jcaiT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi))+disjcaT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi),p.ocopt.s1ho);
            p.hist.vv=[p.hist.vv val];
            p=errcheck(p,sol1);
            if i==2 % if no secant is computed yet
                p.ocopt.um1=p.cp.u;
                p.ocopt.parm1=p.cp.par;
            else
                p.ocopt.um2=p.ocopt.um1;
                p.ocopt.um1=p.cp.u;
                p.ocopt.parm2=p.ocopt.parm1;
                p.ocopt.parm1=p.cp.par;
            end
            if size(p.ocopt.um1,2)~=size(p.ocopt.um2,2)
                p.ocopt.um2=(interp1(told,p.ocopt.um2',p.cp.t))';
            end
        end
        nstart=mst+1;
    else
        nstart=1; % secant given at startup
    end
    if p.ocopt.msw==1
        if length(p.hist.alpha)>1
            sig=p.hist.alpha(end)-p.hist.alpha(end-1);
        else
            sig=al; % if secant is given no previous alpha is avaiable, thus it's expected the secant is from alpha to zero
        end
        p.ocopt.usec=p.ocopt.um1-p.ocopt.um2;
        p.ocopt.usec=p.ocopt.usec/sig;
        p.ocopt.parsec=(p.ocopt.parm1-p.ocopt.parm2)/sig;
        for i=nstart:length(alvin)
            al=alvin(i);
            if p.ocopt.s1ho==1
                p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.hopf.y(1:n,end-1); % start pointalvin(i);
            else
                p.ocopt.u0=al*p.ocopt.s0.u(1:n)+(1-al)*p.ocopt.s1.u(1:n); % start pointalvin(i);
            end
            if ~isempty(p.hist.alpha)
                sig=al-p.hist.alpha(end);
            else
                sig=al; % if secant is given no previous alpha is avaiable, thus it's expected the secant is from alpha to zero
            end
            p.cp.u=p.cp.u+sig*p.ocopt.usec;
            p.cp.par=p.cp.par+sig*p.ocopt.parsec;
            switch  p.ocopt.verb; 
                case 1; fprintf('alpha=%g\n',al);
                case 2; fprintf('alpha=%g - Starting computation...\n',al);
            end
            told=p.cp.t;
            [sol1,p.ocopt.s1]=cpsolver(p.cp,p,0);
            if sol1.err~=0
                warning('Did not find any solution in %i continuation-step, error-code: %i. \n Last solution in p.cp.',i,sol1.err);
                p.cp=sol1;
                return;
            end
            if p.ocopt.retsw==1
                p.hist.tt=[p.hist.tt,sol1.t];
                p.hist.u=[p.hist.u,sol1.u];
                p.hist.par=[p.hist.par,sol1.par];
            else
                p.hist.tt={sol1.t};
            end
            p.hist.alpha=[p.hist.alpha al];
            val=jcaiT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi))+disjcaT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi),p.ocopt.s1ho);
            p.hist.vv=[p.hist.vv val];
            p=errcheck(p,sol1);
            p.ocopt.um2=p.ocopt.um1;
            p.ocopt.um1=p.cp.u;
            if size(p.ocopt.um1,2)~=size(p.ocopt.um2,2)
                p.ocopt.um2=(interp1(told,p.ocopt.um2',p.cp.t))';
            end
            p.ocopt.parm2=p.ocopt.parm1;
            p.ocopt.parm1=p.cp.par;
            p.ocopt.usec=p.ocopt.um1-p.ocopt.um2;
            p.ocopt.usec=p.ocopt.usec/sig;
            p.ocopt.parsec=p.ocopt.parm1-p.ocopt.parm2;
            p.ocopt.parsec=p.ocopt.parsec/sig;
        end
    end
end
if nargin>2 % start arclength continuation
    p.cp.par(2)=p.hist.alpha(end); % add alpha as parameter
    p.ocopt.parm1(2)=p.hist.alpha(end);
    p.ocopt.parm2(2)=p.hist.alpha(end-1);
    p.cp.par=p.cp.par(:);
    p.ocopt.parm1=p.ocopt.parm1(:);
    p.ocopt.parm2=p.ocopt.parm2(:);
    p.ocopt.parsec=p.ocopt.parm1-p.ocopt.parm2;
    % renormate secant for arclength conti
    p.ocopt.usec=p.ocopt.um1-p.ocopt.um2;
    sparnorm=sqrt(vparscal(p.ocopt.usec,p.ocopt.parsec,p.ocopt.usec,p.ocopt.parsec,1/n,1));
    p.ocopt.usec=p.ocopt.usec/sparnorm;
    p.ocopt.parsec=p.ocopt.parsec/sparnorm;
    for i=1:varargin{1}
        ok=0;
        fprintf('i=%i, sig=%g\n',i,p.ocopt.sig);
        if p.ocopt.sig<p.ocopt.sigmin
            error('sig<sigmin, check preferences.');
        end
        while (p.ocopt.sig>=p.ocopt.sigmin) && (ok==0)
            p.cp.u=p.ocopt.um1+p.ocopt.sig*p.ocopt.usec; % new predictor
            p.cp.par=p.ocopt.parm1+p.ocopt.sig*p.ocopt.parsec;
            if p.ocopt.sig>p.ocopt.sigmin
                warning('off','all');
            end
            told=p.cp.t;
            [sol1,p.ocopt.s1]=cpsolver(p.cp,p,1);
            if sol1.err==0
                if p.ocopt.retsw==1
                    p.hist.tt=[p.hist.tt,sol1.t];
                    p.hist.u=[p.hist.u,sol1.u];
                    p.hist.par=[p.hist.par,sol1.par];
                else
                    p.hist.tt={sol1.t};
                end
                ok=1;
                p.hist.alpha=[p.hist.alpha sol1.par(end)];
                val=jcaiT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi))+disjcaT(p.ocopt.s1,sol1,p.ocopt.s1.u(p.ocopt.s1.nu+p.ocopt.rhoi),p.ocopt.s1ho);
                p.hist.vv=[p.hist.vv val];
                p=errcheck(p,sol1);
                p.ocopt.um2=p.ocopt.um1;
                p.ocopt.um1=p.cp.u;
                if length(p.ocopt.um1)~=length(p.ocopt.um2)
                    p.ocopt.um2=interp1(told,p.ocopt.um2',p.cp.t)';
                end
                p.ocopt.usec=p.ocopt.um1-p.ocopt.um2;
                p.ocopt.parm2=p.ocopt.parm1;
                p.ocopt.parm1=p.cp.par;
                p.ocopt.parsec=p.ocopt.parm1-p.ocopt.parm2;
                sparnorm=sqrt(vparscal(p.ocopt.usec,p.ocopt.parsec,p.ocopt.usec,p.ocopt.parsec,1/n,1));
                p.ocopt.usec=p.ocopt.usec/sparnorm;
                p.ocopt.parsec=p.ocopt.parsec/sparnorm;
                fprintf('Arclength continuation converged, new alpha=%g. \n',p.cp.par(end));
                if p.cp.par(2)>1-1e-6
                    fprintf('alpha=%g - abort arc continuation.\n',p.cp.par(2));
                    fprintf('Starting nat continuation to alpha=1. \n');
                    p=isc(p,1);
                    return;
                end
                if (sol1.info.itnl<2 && p.ocopt.sig<p.ocopt.sigmax/1.1)
                    p.ocopt.sig=p.ocopt.sig*1.1;
                end % increase sig
            else
                if p.ocopt.sig>p.ocopt.sigmin
                    p.ocopt.sig=max(p.ocopt.sig/2,p.ocopt.sigmin);
                    fprintf('no convergence, reducing sig to %g\n',p.ocopt.sig);
                else
                    fprintf('no convergence and sig=sigmin \n');
                    warning('Continuation did not converge, change p.ocopt.sigmin or tomopts or use iscnat.');
                    return;
                end
            end
        end
    end
end
end

function p=errcheck(p,sol1)
% computes the deviation from the aimed CPS/CSS and takes appropiated actions.
% verbosity triggered by p.ocopt.verb
n=p.ocopt.s1.nu+p.ocopt.s1.nc.nq;
errmax=max(abs(sol1.u(:,end)-p.ocopt.u1(1:n))); % calculates inf-norm of deviation from CPS/CSS
switch p.ocopt.verb;
    case 2; fprintf('Current maximal deviation from CPS/CSS: %g.\n',errmax);
end
sw=1;
while errmax>10*p.ocopt.derr
    if sw==1 % first loop
        fprintf('Deviation from CPS/CSS is more then 10 times higher then derr.\n');
        sw=0;
    else
        fprintf('Current maximal deviation from CPS/CSS: %g.\n',errmax);
    end
    if p.ocopt.s1ho==1
        % add sufficient numbers of periods
        smuv=abs(p.ocopt.muv(abs(p.ocopt.muv)<1-1e-4)); % select stable floquet multipliers
        lsfm=max(smuv); % select larges stable floquet multiplier
        wTn=max(ceil(log(p.ocopt.derr/errmax)/log(lsfm)),1); % approx number of periods needed to reach derr
        fprintf('Adding %i extra periods. \n',wTn);
        mppp=ceil(length(sol1.t)/p.ocopt.wT); % mesh points per period
        horesc=linspace(0,1,mppp); % rescaled cps time-orbit
        horescd=(interp1(p.ocopt.s1.hopf.t,p.ocopt.s1.hopf.y',horesc))'; % rescaled cps data-orbit
        ttemp=p.ocopt.s1.hopf.T*horesc(2:end)+p.ocopt.s1.hopf.T*(0:wTn-1)'+sol1.par(1);
        ttemp=ttemp';
        sol1.t=1/(p.ocopt.s1.hopf.T*wTn+sol1.par(1))*[sol1.par(1)*sol1.t,ttemp(:)'];
        sol1.par(1)=p.ocopt.s1.hopf.T*wTn+sol1.par(1);
        sol1.u=[sol1.u,repmat(horescd(:,2:end),1,wTn)];
        p.ocopt.wT=p.ocopt.wT+wTn;
    else
        p.ocopt.err=sqrt(norm(p.ocopt.u1-sol1.u(:,end),2)^2/size(sol1.u,1));
        p.ocopt.freeT=1;
        p.ocopt.lastderr=p.ocopt.derr;
        p.ocopt.derr=inf;
    end
    [sol1,p.ocopt.s1]=cpsolver(sol1,p,0);
    if sol1.err~=0
        warning('Did not find any solution while adding periods to control deviation from CPS, error-code: %i. \n One can try to restart computation with more periods. \n Last solution in p.cp.',i,sol1.err);
        p.cp=sol1;
        return;
    end
    if p.ocopt.retsw==1
        p.hist.tt=[p.hist.tt,sol1.t];
        p.hist.u=[p.hist.u,sol1.u];
        p.hist.par=[p.hist.par,sol1.par];
    else
        p.hist.tt={sol1.t};
    end
    errmax=max(abs(sol1.u(:,end)-p.ocopt.u1(1:n)));
end
if p.ocopt.verb>1; fprintf('Current maximal deviation from CPS/CSS: %g.\n',errmax); end 
p.cp=sol1;
end